#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: astrometrynet.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import pathlib
import subprocess
import tempfile
import time
import warnings
from dataclasses import dataclass
from glob import glob

from typing import NamedTuple, Optional, TypeVar

import numpy
import pandas
from astropy.io import fits
from astropy.table import Table
from astropy.units import UnitsWarning
from astropy.wcs import WCS, FITSFixedWarning


PathLike = TypeVar("PathLike", pathlib.Path, str)


class TimedProcess(NamedTuple):
    """A completed process which includes its elapsed time."""

    process: asyncio.subprocess.Process | subprocess.CompletedProcess
    elapsed: float


class AstrometryNet:
    """A wrapper for the astrometry.net ``solve-field`` command.

    Parameters
    ----------
    configure_params
        Parameters to be passed to `.configure`.
    """

    def __init__(self, **configure_params):
        solve_field_cmd = subprocess.run(
            "which solve-field",
            shell=True,
            capture_output=True,
        )
        solve_field_cmd.check_returncode()

        self.solve_field_cmd = solve_field_cmd.stdout.decode().strip()

        self._options = {}
        self.configure(**configure_params)

    def configure(
        self,
        backend_config: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        sort_column: Optional[str] = None,
        sort_ascending: Optional[bool] = None,
        no_plots: Optional[bool] = None,
        ra: Optional[float] = None,
        dec: Optional[float] = None,
        radius: Optional[float] = None,
        scale_low: Optional[float] = None,
        scale_high: Optional[float] = None,
        scale_units: Optional[str] = None,
        dir: Optional[str] = None,
        **kwargs,
    ):
        """Configures how to run of ``solve-field```.

        The parameters this method accepts are identical to those of
        ``solve-field`` and are passed unchanged.

        Parameters
        ----------
        backend_config
            Use this config file for the ``astrometry-engine`` program.
        width
            Specify the field width, in pixels.
        height
            Specify the field height, in pixels.
        sort_column
            The FITS column that should be used to sort the sources.
        sort_ascending
            Sort in ascending order (smallest first);
            default is descending order.
        no_plot
            Do not produce plots.
        ra
            RA of field center for search, in degrees.
        dec
            Dec of field center for search, in degrees.
        radius
            Only search in indexes within ``radius`` degrees of the field
            center given by ``ra`` and ``dec``.
        scale_low
            Lower bound of image scale estimate.
        scale_high
            Upper bound of image scale estimate.
        scale_units
            In what units are the lower and upper bounds? Choices:
            ``'degwidth'``, ``'arcminwidth'``, ``'arcsecperpix'``,
            ``'focalmm'``.
        dir
            Path to the directory where all output files will be saved.
        """

        self._options = {
            "backend-config": backend_config,
            "width": width,
            "height": height,
            "sort-column": sort_column,
            "sort-ascending": sort_ascending,
            "no-plots": no_plots,
            "ra": ra,
            "dec": dec,
            "radius": radius,
            "scale-low": scale_low,
            "scale-high": scale_high,
            "scale-units": scale_units,
            "dir": dir,
            "overwrite": True,
            "no-tweak": True,
            "tweak-order": 0,
        }
        self._options.update(self._format_options(kwargs))

        return

    def _format_options(self, options: dict):
        """Formats the option names."""

        formatted_options = {}

        for key, value in options.items():
            formatted_key = key.replace("_", "-")
            formatted_options[formatted_key] = value

        return formatted_options

    def _build_command(self, files, options=None):
        """Builds the ``solve-field`` command to run."""

        if options is None:
            options = self._options

        flags = ["no-plots", "sort-ascending", "overwrite", "no-tweak"]

        cmd = [self.solve_field_cmd]

        for option in options:
            if options[option] is None:
                continue
            if option in flags:
                if options[option] is True:
                    cmd.append("--" + option)
            else:
                cmd.append("--" + option)
                cmd.append(str(options[option]))

        cmd += list(files)

        return cmd

    def run_sync(
        self,
        files: PathLike | list[PathLike],
        shell: bool = True,
        stdout: Optional[PathLike] = None,
        stderr: Optional[PathLike] = None,
        **kwargs,
    ):
        """Runs astrometry.net synchronously.

        Parameters
        ----------
        files
            List of files to be processed.
        shell
            Whether to call `subprocess.run` with ``shell=True``.
        stdout
            Path where to save the stdout output.
        stderr
            Path where to save the stderr output.
        kwargs
            Configuration parameters (see `.configure`) to override. The
            configuration applies only to this run of ``solve-field`` and it
            is not saved.

        Returns
        -------
        `subprocess.CompletedProcess`
            The completed process.

        """

        options = self._options.copy()
        options.update(self._format_options(kwargs))

        if not isinstance(files, (tuple, list)):
            files = [files]

        t0 = time.time()

        args = self._build_command(list(map(str, files)), options=options)
        if shell:
            args = " ".join(args)

        cmd = subprocess.run(
            args,
            shell=shell,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        elapsed = time.time() - t0

        if stdout:
            with open(stdout, "wb") as out:
                if isinstance(args, str):
                    out.write(args.encode() + b"\n")
                else:
                    out.write(" ".join(args).encode() + b"\n")
                out.write(cmd.stdout)

        if stderr:
            with open(stderr, "wb") as err:
                err.write(cmd.stderr)

        return TimedProcess(cmd, elapsed)

    async def run_async(
        self,
        files: PathLike | list[PathLike],
        shell: bool = True,
        stdout: Optional[PathLike] = None,
        stderr: Optional[PathLike] = None,
        **kwargs,
    ) -> TimedProcess:
        """Runs astrometry.net asynchronously.

        Parameters
        ----------
        files
            List of files to be processed.
        shell
            Whether to call `subprocess.run` with ``shell=True``.
        stdout
            Path where to save the stdout output.
        stderr
            Path where to save the stderr output.
        kwargs
            Configuration parameters (see `.configure`) to override. The
            configuration applies only to this run of ``solve-field`` and it
            is not saved.

        Returns
        -------
        `subprocess.CompletedProcess`
            The completed process.

        """

        options = self._options.copy()
        options.update(self._format_options(kwargs))

        if not isinstance(files, (tuple, list)):
            files = [files]

        t0 = time.time()

        args = self._build_command(list(map(str, files)), options=options)

        if shell:
            cmd = await asyncio.create_subprocess_shell(
                " ".join(args),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

        else:
            cmd = await asyncio.create_subprocess_exec(
                *args,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

        stdout_bytes, stderr_bytes = await cmd.communicate()

        elapsed = time.time() - t0

        if stdout:
            with open(stdout, "wb") as out:
                if isinstance(args, str):
                    out.write(args.encode() + b"\n")
                else:
                    out.write(" ".join(args).encode() + b"\n")
                out.write(stdout_bytes)

        if stderr:
            with open(stderr, "wb") as err:
                err.write(stderr_bytes)

        return TimedProcess(cmd, elapsed)


def astrometrynet_quick(
    index_paths: dict[int, str],
    regions: pandas.DataFrame,
    ra: float,
    dec: float,
    pixel_scale: float,
    pixel_scale_factor_hi: float = 1.1,
    pixel_scale_factor_lo: float = 0.9,
    output_root: str | None = None,
    scales: dict[int, list[int]] | None = None,
    radius: float = 0.5,
    width: float | None = None,
    height: float | None = None,
    plot: bool = False,
    cpulimit: int = 30,
    raise_on_unsolved: bool = False,
    **kwargs,
) -> AstrometrySolution:
    """Quickly process a set of detections using astrometry.net.

    Parameters
    ----------
    index_paths
        The paths to the index files to use. A dictionary of series number
        to path on disk, e.g.,
        ``{5200: "/data/astrometrynet/5200", 4100: "/data/astrometrynet/5100"}``.
    regions
        A pandas data frame of source detections with at least three columns:
        ``x``, ``y``, and ``flux``. ``flux`` can be set to all ones if the
        fluxes have not been measured.
    ra
        The estimated RA of the field, in degrees.
    dec
        The estimated declination of the field, in degrees.
    pixel_scale
        The estimated pixel scale, in arcsec.
    pixel_scale_factor_lo
        A multiplicative factor for the pixel scale to calculate the
        lowest possible pixel scale to attempt.
    pixel_scale_factor_hi
        A multiplicative factor for the pixel scale to calculate the
        highest possible pixel scale to attempt.
    output_root
        The path where to save astrometry.net file. Must include the directory
        and a root after which input and output files will be named. If `None`,
        the xyls file is saved to a temporary file.
    scales
        Index scales to use. Otherwise uses all index files. The format
        is a dictionary of series number to a list of scales, e.g.,
        ``{5200: [4, 5], 4100: [10, 11]}``.
    radius
        The radius, in degrees, around ``ra`` and ``dec`` to search.
    width
        The width of the image in pixels.
    height
        The height of the image in pixels.
    plot
        Whether to have astrometry.net generate plots.
    cpulimit
        Maximum time to wait for a solution, in seconds.
    raise_on_unsolved
        Raise an error if the field is not solved.
    kwargs
        Other options to pass to `.AstrometryNet`.

    Returns
    -------
    solution
        An `.AstrometrySolution` object with the astrometric solution.

    """

    if "flux" not in regions:
        regions.loc[:, "flux"] = 1

    xyls = Table.from_pandas(regions)

    if output_root:
        output_root = os.path.abspath(output_root)
    else:
        output_root = tempfile.NamedTemporaryFile().name
        plot = False

    dirname = os.path.dirname(output_root)
    outfile = os.path.basename(output_root)

    corr_file = os.path.join(dirname, outfile + ".corr")

    if not os.path.exists(dirname):
        os.makedirs(dirname, exist_ok=True)

    backend_config = os.path.join(dirname, f"{outfile}.cfg")
    with open(backend_config, "w") as ff:
        ff.write("inparallel\n")
        ff.write(f"cpulimit {cpulimit}\n")

        for index_path in index_paths.values():
            ff.write(f"add_path {os.path.abspath(index_path)}\n")

        for series, index_path in index_paths.items():
            index_path = os.path.abspath(index_path)
            this_scales = scales.get(series, None) if scales is not None else None
            if this_scales is None:
                scale_files = glob("index-*", root_dir=index_path)
                for scale_file in scale_files:
                    ff.write(f"index {scale_file}\n")
            else:
                for this_scale in this_scales:
                    series_scale = series + this_scale
                    scale_files = glob(f"index-{series_scale}*", root_dir=index_path)
                    if len(scale_files) == 0:
                        warnings.warn(f"No indices found for scale {series_scale}.")
                    for scale_file in scale_files:
                        ff.write(f"index {scale_file}\n")

    opts = dict(
        backend_config=str(backend_config),
        width=width or max(regions.loc[:, "x"]),
        height=height or max(regions.loc[:, "y"]),
        no_plots=not plot,
        scale_low=pixel_scale * pixel_scale_factor_lo,
        scale_high=pixel_scale * pixel_scale_factor_hi,
        scale_units="arcsecperpix",
        sort_column="flux",
        radius=radius,
        corr=corr_file,
    )
    opts.update(kwargs)

    astrometry = AstrometryNet(**opts)

    xyls_path = os.path.join(dirname, outfile + ".xyls")
    hdus = fits.HDUList(
        [fits.PrimaryHDU(), fits.BinTableHDU(data=xyls.as_array())]  # type:ignore
    )
    hdus.writeto(xyls_path, overwrite=True)

    wcs_output = pathlib.Path(dirname) / (outfile + ".wcs")
    wcs_output.unlink(missing_ok=True)

    stdout = os.path.join(dirname, outfile + ".stdout")
    stderr = os.path.join(dirname, outfile + ".stderr")
    astrometry.run_sync(
        [xyls_path],
        stdout=stdout,
        stderr=stderr,
        ra=ra,
        dec=dec,
    )

    stdout_s = open(stdout).read() if os.path.exists(stdout) else ""
    stderr_s = open(stderr).read() if os.path.exists(stderr) else ""

    solution = AstrometrySolution(wcs=None, xyls=xyls.to_pandas(), stdout="", stderr="")

    if not wcs_output.exists():
        if raise_on_unsolved:
            if stdout_s != "":
                raise RuntimeError(stdout_s + "\n" + stderr_s)
            else:
                raise RuntimeError("Unexpected error. No stderr was generated.")
        else:
            return solution

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=FITSFixedWarning)
        wcs = WCS(open(wcs_output).read(), relax=True)

    solution.wcs = wcs
    solution.stdout = open(stdout).read()
    solution.stderr = open(stderr).read()

    if os.path.exists(corr_file):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UnitsWarning)

            stars = Table.read(corr_file).to_pandas()
            solution.stars = stars

    return solution


@dataclass
class AstrometrySolution:
    """An astrometric solution."""

    wcs: WCS | None
    xyls: pandas.DataFrame
    stdout: str
    stderr: str
    stars: pandas.DataFrame | None = None

    def __repr__(self) -> str:
        return f"<AstrometrySolution (solved={self.solved})>"

    @property
    def solved(self):
        """Whether a solution was found."""

        return self.wcs is not None

    @property
    def transformation_matrix(self):
        """Returns the transformation matrix."""

        if not self.wcs:
            raise RuntimeError("No solution found.")

        return self.wcs.wcs.cd

    @property
    def rotation(self):
        """Returns the rotation angle."""

        if not self.wcs:
            raise RuntimeError("No solution found.")

        tm = self.transformation_matrix
        return numpy.degrees(numpy.arctan2(tm[0, 0], tm[1, 0]))

    @property
    def pixel_scale(self):
        """Returns the pixel scale."""

        if not self.wcs:
            raise RuntimeError("No solution found.")

        tm = self.transformation_matrix

        return numpy.abs(tm[0, 0] * 3600)

    def pixel_to_world(self, x: float, y: float):
        """Returns the sky coordinate for a pixel position."""

        if not self.wcs:
            raise RuntimeError("No solution found.")

        return self.wcs.pixel_to_world(x, y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: focus.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy
import pandas
import seaborn
from astropy.io import fits
from scipy.interpolate import UnivariateSpline

from sdsstools.logger import get_logger

from lvmguider.tools import run_in_executor


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


def fit_parabola(x_arr: list[float], y_arr: list[float], y_err: list[float]):
    """Fits a parabola to the data."""

    w = 1 / numpy.array(y_err)
    a, b, c = numpy.polyfit(x_arr, y_arr, 2, w=w, full=False)

    corr_matrix = numpy.corrcoef(y_arr, numpy.polyval([a, b, c], x_arr))
    corr = corr_matrix[0, 1]
    R_sq = corr**2

    return (a, b, c, R_sq)


def fit_spline(x: numpy.ndarray, y: numpy.ndarray, w: numpy.ndarray | None = None):
    """Fits a spline to data."""

    spl = UnivariateSpline(x, y)

    corr_matrix = numpy.corrcoef(y, spl(x))
    R2 = corr_matrix[0, 1] ** 2

    return spl, R2


class Focuser:
    """Focus a telescope.

    Parameters
    ----------
    telescope
        The telescope being focused: ``'sci'``, ``'spec'``, ``'skye'``,
        or ``'skyw'``.

    """

    def __init__(self, telescope: str):
        self.telescope = telescope
        self.foc_actor = f"lvm.{self.telescope}.foc"  # Focuser

    async def focus(
        self,
        command: GuiderCommand,
        initial_guess: float | None = None,
        step_size: float = 0.5,
        steps: int = 7,
        exposure_time: float = 5.0,
        fit_method="parabola",
        plot: bool = True,
        plot_dir: str = "qa/focus",
    ) -> tuple[pandas.DataFrame, dict]:
        """Performs the focus routine.

        Parameters
        ----------
        command
            The actor command to use to talk to other actors.
        initial_guess
            An initial guess of the focus position, in DT.
        step_size
            The resolution of the focus sweep in DT steps.
        steps
            Number of steps in the focus sweep. Must be an odd number.
        exposure_time
            The exposure time for each AG frame.
        fit_method
            The method used to fit the data. Either ``'parabola'`` for a second
            order polynomial fit, or ``'spline'`` to fit a univariate spline
            to the data.
        plot
            Whether to produce focus plots. The plots are saved asynchronously
            and this method may return before the plots are available.
        plot_dir
            The path where to save the focus plot. Always relative to the
            path of the AG frames. The plot will be saved as
            ``{plot_dir}/focus_{telescope}_{frame0}_{frame1}.pdf`` where
            ``{frame0}`` and ``{frame1}`` are the first and last frame number
            of the sweep sequence.

        """

        cameras = command.actor.cameras

        steps = int(steps)
        if steps % 2 == 0:
            steps += 1

        if initial_guess is None:
            cmd = await command.send_command(self.foc_actor, "getPosition")
            if cmd.status.did_fail:
                raise RuntimeError("Failed retrieving position from focuser.")
            initial_guess = cmd.replies.get("Position")
            command.debug(f"Using focuser position: {initial_guess} DT")

        assert initial_guess is not None

        focus_grid = numpy.arange(
            initial_guess - (steps // 2) * step_size,
            initial_guess + (steps // 2 + 1) * step_size,
            step_size,
        )

        if numpy.any(focus_grid <= 0):
            raise ValueError("Focus values out of range.")

        source_list: list[pandas.DataFrame] = []
        files: list[pathlib.Path] = []
        framenos: list[int] = []

        for focus_position in focus_grid:
            await self.goto_focus_position(command, focus_position)

            step_files, frameno, sources = await cameras.expose(
                command,
                exposure_time=exposure_time,
                extract_sources=True,
            )

            files += step_files
            framenos.append(frameno)

            if sources is None or len(sources) == 0:
                command.warning(f"No sources detected at {focus_position} DT.")
                continue

            asources = pandas.concat(sources)
            if len(asources) == 0:
                command.warning(
                    f"No sources found for focus position {focus_position} DT. "
                    "Skipping this point."
                )
                continue

            asources["dt"] = focus_position

            fwhm = float(asources["fwhm"].median())
            command.info(
                focus_point=dict(
                    focus=round(focus_position, 2),
                    n_sources=len(asources),
                    fwhm=round(fwhm, 2),
                )
            )

            source_list.append(asources)

        if len(source_list) < 3:
            raise ValueError("Insufficient number of focus points.")

        sources = pandas.concat(source_list)

        fit_data = self.fit_focus(
            sources,
            fit_method=fit_method,
        )

        command.info(
            best_focus=dict(
                focus=numpy.round(fit_data["xmin"], 2),
                fwhm=numpy.round(fit_data["ymin"], 2),
                r2=numpy.round(fit_data["R2"], 3),
            )
        )

        if plot:
            frame0 = min(framenos)
            frame1 = max(framenos)
            basepath = files[-1].absolute().parent

            plot_name = f"focus_{self.telescope}_{frame0}_{frame1}.pdf"
            plot_path = basepath / plot_dir / plot_name

            asyncio.create_task(
                run_in_executor(
                    self.plot,
                    sources,
                    fit_method,
                    plot_path,
                    fit_params=fit_data,
                )
            )

        await self.goto_focus_position(command, numpy.round(fit_data["xmin"], 2))

        return sources, fit_data

    async def goto_focus_position(self, command: GuiderCommand, focus_position: float):
        """Moves the focuser to a position."""

        cmd = await command.send_command(
            self.foc_actor,
            f"moveAbsolute {focus_position} DT",
        )
        if cmd.status.did_fail:
            raise RuntimeError(f"Failed reaching focus {focus_position:.2f} DT.")
        if cmd.replies.get("AtLimit") is True:
            raise RuntimeError("Hit a limit while focusing.")

    def fit_focus(self, sources: pandas.DataFrame, fit_method: str = "parabola"):
        """Fits the data and returns the best focus and measured FWHM."""

        sources_valid = sources.loc[(sources.xfitvalid == 1) & (sources.yfitvalid == 1)]

        if fit_method == "parabola":
            a, b, c, R2 = fit_parabola(
                sources_valid.dt.tolist(),
                sources_valid.fwhm.tolist(),
                sources_valid.xrms.tolist(),
            )

            xmin = -b / 2 / a
            ymin = a * xmin**2 + b * xmin + c

            return {"xmin": xmin, "ymin": ymin, "R2": R2, "coeffs": [a, b, c]}

        elif fit_method == "spline":
            fwhm = sources_valid.groupby("dt").apply(
                lambda gg: pandas.Series(
                    {
                        "fwhm": gg.fwhm.median(),
                        "std": gg.fwhm.median(),
                    }
                )
            )

            x = fwhm.index.to_numpy()
            y = fwhm.fwhm.to_numpy()
            w = 1.0 / fwhm["std"].to_numpy()
            spline, R2 = fit_spline(x, y, w=w)

            x_refine = numpy.arange(numpy.min(x), numpy.max(x), 0.01)
            y_refine = spline(x_refine)

            arg_min = numpy.argmin(y_refine)
            xmin = x_refine[arg_min]
            ymin = y_refine[arg_min]

            return {"xmin": xmin, "ymin": ymin, "R2": R2, "spline": spline}

        else:
            raise ValueError("Invalid fit method.")

    def plot(
        self,
        data: pandas.DataFrame,
        fit_method: str,
        plot_path: pathlib.Path,
        fit_params: dict = {},
    ):
        """Produces a plot for the focus sequence."""

        seaborn.set_theme(palette="deep")

        data_valid = data.loc[(data.xfitvalid == 1) & (data.yfitvalid == 1)]

        with plt.ioff():
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(13, 8))

            ax.scatter(
                data_valid.dt,
                data_valid.fwhm,
                s=10,
                color="y",
                ec="None",
                alpha=0.3,
            )

            fwhm_median = data_valid.groupby("dt").fwhm.median()

            ax.scatter(
                fwhm_median.index,
                fwhm_median,
                ec="None",
                color="r",
                alpha=0.8,
                s=20,
            )

            xmin = numpy.min(data_valid.dt.to_numpy()).astype(numpy.float32)
            xmax = numpy.max(data_valid.dt.to_numpy()).astype(numpy.float32)

            xx = numpy.arange(xmin - 0.5, xmax + 0.5, 0.01)

            if fit_method == "parabola":
                a, b, c = fit_params["coeffs"]
                yy = a * xx**2 + b * xx + c

            elif fit_method == "spline":
                spline = fit_params["spline"]
                yy = spline(xx)

            else:
                raise ValueError("Invalid fit_method.")

            ax.plot(xx, yy, "m-", zorder=20)

            ax.axvline(
                x=fit_params["xmin"],
                linestyle="dashed",
                color="k",
                alpha=0.5,
                zorder=10,
            )

            ax.set_ylim(0.5, max(fwhm_median.values.astype(float)) + 0.5)
            ax.set_xlabel("Focuser position [DT]")
            ax.set_ylabel("FWHM [arcsec]")

            ax.set_title(
                f"Best focus: {fit_params['xmin']:.2f} DT; "
                f"FWHM: {fit_params['ymin']:.2f} arcsec"
            )

            plot_path.parent.mkdir(parents=True, exist_ok=True)
            fig.savefig(str(plot_path))

            plt.close(fig)

        seaborn.reset_defaults()

    def reprocess(
        self,
        files: list[str] | list[pathlib.Path],
        fit_method="parabola",
        plot: bool = True,
        plot_dir: str = "qa/focus",
    ):
        """Reprocesses a list of files.

        Parameters
        ----------
        files
            A list of files to reprocess. They must correspond to a focus
            sweep taken using this telescope.
        fit_method
            The method used to fit the data. Either ``'parabola'`` for a second
            order polynomial fit, or ``'spline'`` to fit a univariate spline
            to the data.
        plot
            Whether to produce focus plots.
        plot_dir
            The path where to save the focus plot. Always relative to the
            path of the AG frames. The plot will be saved as
            ``{plot_dir}/focus_{telescope}_{frame0}_{frame1}.pdf`` where
            ``{frame0}`` and ``{frame1}`` are the first and last frame number
            of the sweep sequence.

        """

        log = get_logger(f"lvmguider.{self.telescope}.reprocess", use_rich_handler=True)

        log.info(f"Reprocessing {len(files)} files.")

        sources = []
        framenos = []

        for file in files:
            framenos.append(int(str(file).split(".")[-2].split("_")[1]))

            hdul = fits.open(str(file))
            focus_position = hdul["RAW"].header["FOCUSDT"]

            sources.append(pandas.DataFrame(hdul["SOURCES"].data))
            sources[-1]["dt"] = focus_position

        data = pandas.concat(sources)

        fit_data = self.fit_focus(
            data,
            fit_method=fit_method,
        )

        log.info(
            f"Best focus: {fit_data['xmin']:.2f} DT; "
            f"FWHM: {fit_data['ymin']: .2f} arcsec; "
            f"R2: {fit_data['R2']:.3f}"
        )

        if plot:
            frame0 = min(framenos)
            frame1 = max(framenos)
            basepath = pathlib.Path(files[-1]).absolute().parent

            plot_name = f"focus_{self.telescope}_{frame0}_{frame1}.pdf"
            plot_path = basepath / plot_dir / plot_name

            self.plot(data, fit_method, plot_path, fit_params=fit_data)
            log.info(f"QA plot saved to {plot_path!s}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-23
# @Filename: coadd.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import multiprocessing
import os
import os.path
import pathlib
import re
import time
from functools import partial
from threading import Lock

from typing import TYPE_CHECKING, Literal, Sequence

import numpy
import pandas
from astropy.io import fits
from astropy.stats import sigma_clip
from astropy.time import Time
from packaging.version import Version
from watchdog.events import FileCreatedEvent, PatternMatchingEventHandler
from watchdog.observers.polling import PollingObserver

from sdsstools.time import get_sjd

from lvmguider import config, log
from lvmguider.dataclasses import (
    CameraSolution,
    CoAdd_CameraSolution,
    FrameData,
    GlobalSolution,
    GuiderSolution,
)
from lvmguider.extraction import extract_marginal, extract_sources
from lvmguider.plotting import plot_qa
from lvmguider.tools import (
    angle_difference,
    dataframe_to_database,
    estimate_zeropoint,
    get_dark_subtracted_data,
    get_frameno,
    get_guider_files_from_spec,
    get_raw_extension,
    get_spec_frameno,
    header_from_model,
    isot_to_sjd,
    nan_or_none,
    polyfit_with_sigclip,
    sort_files_by_camera,
)
from lvmguider.transformations import (
    ag_to_full_frame,
    get_crota2,
    match_with_gaia,
    solve_camera_with_astrometrynet,
    wcs_from_gaia,
)
from lvmguider.types import ARRAY_2D_F32


if TYPE_CHECKING:
    from lvmguider.astrometrynet import AstrometrySolution


AnyPath = str | os.PathLike

MULTIPROCESS_MODE_TYPE = Literal["frames"] | Literal["cameras"] | Literal["telescopes"]
MULTIPROCESS_NCORES: dict[str, int] = {"frames": 4, "cameras": 2, "telescopes": 4}

TELESCOPES: list[str] = ["sci", "spec", "skye", "skyw"]


def process_all_spec_frames(
    path: AnyPath,
    create_summary: bool = True,
    write_to_database: bool = True,
    multiprocess_mode: MULTIPROCESS_MODE_TYPE | None = "telescopes",
    **kwargs,
):
    """Processes all the spectrograph frames in a directory.

    Parameters
    ----------
    path
        The path to the directory to process. This is usually a
        ``/data/spectro/<MJD>`` directory.
    create_summary
        Generates a single table file with all the guider data for
        the processed files.
    write_to_database
        Whether to load the co-added data to the database.
    multiprocess_mode
        Defines how to use parallel processing when processing multiple
        images. Options are ``telescopes`` which will process each telescope
        for a spectrograph image in parallel, ``cameras``, which will process
        each camera in parallel, and ``frames`` which will process frames within
        one camera in parallel. The number of cores used by each one of
        these modes can be adjusted by modifying the `.MULTIPROCESS_NCORES`
        dictionary.
    kwargs
        Keyword arguments to pass to `.coadd_from_spec_frame`.

    """

    path = pathlib.Path(path)
    spec_files = sorted(path.glob("sdR-*-b1-*.fits.gz"))

    if len(spec_files) == 0:
        return

    date_obs = fits.getval(spec_files[0], "OBSTIME")
    sjd = get_sjd("LCO", Time(date_obs, format="isot").to_datetime())

    all_paths: dict[int, list[pathlib.Path]] = {}

    for file in spec_files:
        header = fits.getheader(str(file))
        specno = get_spec_frameno(file)

        # Skip cals and frames without any guider information.
        if header["IMAGETYP"] != "object":
            continue

        if header["EXPTIME"] < 120:
            log.warning(f"Spec frame {file.name!s} is too short. Skipping.")
            continue

        paths = coadd_from_spec_frame(
            file,
            multiprocess_mode=multiprocess_mode,
            write_to_database=write_to_database,
            **kwargs,
        )
        all_paths[specno] = paths

    if create_summary:
        gs_paths: list[pathlib.Path] = []
        spec_nos: list[int] = []

        for spec_no in all_paths:
            gs_paths += all_paths[spec_no]
            spec_nos += [spec_no] * len(all_paths[spec_no])

        g_fs = [pp.with_name(pp.stem + "_guiderdata.parquet") for pp in gs_paths]
        g_fs = [path for path in g_fs if path.exists()]
        if len(g_fs) > 0:
            base_path = g_fs[0].parent
            outpath = base_path / f"lvm.guider.mjd_{sjd}_guiderdata.parquet"
            log.info(f"Saving MJD summary file to {outpath!s}")
            create_summary_file(g_fs, outpath, list(spec_nos))

        f_fs = [pp.with_name(pp.stem + "_frames.parquet") for pp in gs_paths]
        f_fs = [path for path in f_fs if path.exists()]
        if len(f_fs) > 0:
            base_path = f_fs[0].parent
            outpath = base_path / f"lvm.guider.mjd_{sjd}_frames.parquet"
            log.info(f"Saving MJD summary file to {outpath!s}")
            create_summary_file(f_fs, outpath, list(spec_nos))


def coadd_telescope(
    file: pathlib.Path,
    outpath: str | None,
    telescope: str,
    use_time_range: bool = False,
    write_to_database: bool = True,
    **kwargs,
):
    """A helper function that processes one telescopes for a spectrograph image.

    This function is called by `.coadd_from_spec_frame` for a given ``file``
    and ``telescope``. It's split from the code from that function to allow
    parallelisation.

    """

    log.info(
        f"Generating co-added frame for {telescope!r} for "
        f"spectrograph file {file!s}"
    )

    try:
        log.debug("Identifying guider frames.")
        frames = get_guider_files_from_spec(
            file,
            telescope=telescope,
            use_time_range=use_time_range,
        )
    except Exception as err:
        log.error(f"Cannot process {telescope!r} for {file!s}: {err}")
        return

    if len(frames) < 4:
        log.warning(f"No guider frames found for {telescope!r} in {file!s}")
        return

    exposure_match = re.search("[0-9]{8}", file.name)
    exposure_no = int(exposure_match.group(0)) if exposure_match else None

    try:
        gs = create_global_coadd(
            frames,
            telescope=telescope,
            outpath=outpath,
            **kwargs,
        )

        if gs.path is not None:
            if write_to_database and exposure_no is not None:
                coadd_to_database(gs.path, exposure_no=exposure_no)
            return gs.path

    except Exception as err:
        log.critical(
            f"Failed generating co-added frames for {file!s} on "
            f"telescope {telescope!r}: {err}"
        )
        return None


def coadd_from_spec_frame(
    file: AnyPath,
    outpath: str | None = None,
    telescopes: list[str] = TELESCOPES,
    use_time_range: bool | None = None,
    write_to_database: bool = True,
    multiprocess_mode: MULTIPROCESS_MODE_TYPE | None = "telescopes",
    **kwargs,
):
    """Processes the guider frames associated with an spectrograph file.

    Parameters
    ----------
    file
        The spectro file to process. Guider frames will be selected
        from the header keywords in the file.
    output
        The path where to save the global co-added frame. If `None`,
        uses the default path which includes the frame number of the
        spectro file.
    telescopes
        The list of telescope guider frames to process.
    use_time_range
        If `True`, selects guider frames from their timestamps instead
        of using the headers in the spectro file.
    multiprocess_mode
        See `.process_all_spec_frames`.
    kwargs
        Keyword arguments to pass to `.create_global_coadd`.

    """

    outpath = outpath or config["coadds"]["paths"]["coadd_spec_path"]

    file = pathlib.Path(file)
    if not file.exists():
        raise FileExistsError(f"File {file!s} not found.")

    specno = get_spec_frameno(file)
    outpath = outpath.format(specno=specno)

    paths: list[pathlib.Path | None] = []

    coadd_telescope_p = partial(
        coadd_telescope,
        file,
        outpath,
        use_time_range=use_time_range,
        multiprocess_mode=multiprocess_mode,
        write_to_database=write_to_database,
        **kwargs,
    )
    if multiprocess_mode == "telescopes":
        with multiprocessing.Pool(MULTIPROCESS_NCORES["telescopes"]) as pool:
            paths += pool.map(coadd_telescope_p, telescopes)
    else:
        for telescope in telescopes:
            paths.append(coadd_telescope_p(telescope))

    return [path for path in paths if path is not None]


def create_global_coadd(
    files: Sequence[AnyPath],
    telescope: str,
    outpath: str | None = "default",
    save_camera_coadded: bool = False,
    generate_qa: bool = True,
    **coadd_camera_kwargs,
):
    """Produces a global co-added frame.

    Calls `.coadd_camera` with the frames for each camera and produces
    a single co-added file with the extensions from the camera co-added frames
    and an extension ``COADD`` with the global astrometric solution of the
    full frame.

    Parameters
    ----------
    files
        The list of files to co-add. Must include frames from all the cameras for
        a given guider sequence.
    telescope
        The telescope associated with the images.
    outpath
        The path of the global co-added frame. If a relative path, it is written
        relative to the path of the first file in the list. If `False`, the file
        is not written to disk but the ``HDUList` is returned. With ``"default"``,
        the default output path is used.
    save_camera_coadded
        Whether to write to disk the individual co-added images for each camera.
    coadd_camera_kwargs
        Arguments to pass to `.coadd_camera` for each set of camera frames.
        It includes the ``multiprocess_mode`` parameter which if set to ``cameras``
        will process each camera in parallel.

    Returns
    -------
    global_solution
        The `.GlobalSolution` instance with the information about the global
        frame and solution.

    """

    if outpath == "default":
        outpath = config["coadds"]["paths"]["coadd_path"]

    coadd_camera_kwargs = coadd_camera_kwargs.copy()
    if save_camera_coadded is False:
        coadd_camera_kwargs["outpath"] = None

    camera_to_files = sort_files_by_camera(files)
    cameras = [camera for camera in camera_to_files if len(camera_to_files[camera]) > 0]
    camera_files = [camera_to_files[camera] for camera in cameras]

    frame_nos = set([get_frameno(file) for file in files])

    # Co-add each camera independently.
    coadd_camera_partial = partial(coadd_camera, **coadd_camera_kwargs)

    multiprocess_mode = coadd_camera_kwargs.get("multiprocess_mode", None)
    if multiprocess_mode == "cameras":
        with multiprocessing.Pool(MULTIPROCESS_NCORES["cameras"]) as pool:
            coadd_solutions = list(pool.map(coadd_camera_partial, camera_files))
    else:
        coadd_solutions = list(map(coadd_camera_partial, camera_files))

    # Now create a global solution.
    root = pathlib.Path(files[0]).parent

    get_guider_solutions_p = partial(get_guider_solutions, root, telescope)
    if multiprocess_mode == "cameras":
        with multiprocessing.Pool(MULTIPROCESS_NCORES["cameras"]) as pool:
            guider_solutions = list(pool.map(get_guider_solutions_p, list(frame_nos)))
    else:
        guider_solutions = map(get_guider_solutions_p, list(frame_nos))

    gs = GlobalSolution(
        coadd_solutions=coadd_solutions,
        guider_solutions=[gs for gs in guider_solutions if gs is not None],
        telescope=telescope,
    )

    # Fit a new WCS from the individual co-added solutions using the full frame.
    sources = gs.sources
    if telescope != "spec":
        if len(sources.ra.dropna()) > 5:
            gs.wcs = wcs_from_gaia(sources, xy_cols=["x_ff", "z_ff"])
        else:
            log.warning("Unable to fit global WCS. Not enough matched sources.")

    if outpath is not None:
        # Create the path for the output file.
        sample_raw_hdu = get_raw_extension(files[0])
        date_obs = sample_raw_hdu.header["DATE-OBS"]
        sjd = get_sjd("LCO", Time(date_obs, format="isot").to_datetime())

        frameno0 = min(frame_nos)
        frameno1 = max(frame_nos)
        outpath = outpath.format(
            telescope=telescope,
            frameno0=frameno0,
            frameno1=frameno1,
            mjd=sjd,
        )

        if pathlib.Path(outpath).is_absolute():
            path = pathlib.Path(outpath)
        else:
            path = root / outpath

        if path.parent.exists() is False:
            path.parent.mkdir(parents=True, exist_ok=True)

        gs.path = path

        log.info(f"Writing co-added frame to {path.absolute()!s}")

        # Write global header and co-added frames.
        hdul = fits.HDUList(
            [
                fits.PrimaryHDU(),
                fits.ImageHDU(
                    data=None,
                    header=create_global_header(gs),
                    name="GLOBAL",
                ),
            ]
        )

        for coadd in gs.coadd_solutions:
            hdul.append(
                fits.CompImageHDU(
                    data=coadd.coadd_image,
                    header=create_coadd_header(coadd),
                    name=f"COADD_{coadd.camera.upper()}",
                )
            )

        hdul.writeto(str(path), overwrite=True)

        # Write sources to file.
        if gs.sources is not None:
            sources_path = path.with_name(path.stem + "_sources.parquet")
            log.debug(f"Writing co-added sources to {sources_path!s}")
            gs.sources.reset_index(drop=True, inplace=True)
            gs.sources.to_parquet(sources_path)

        # Write guider data.
        guider_data = gs.guider_data()
        guider_path = path.with_name(path.stem + "_guiderdata.parquet")
        log.debug(f"Writing guide data to {guider_path!s}")
        guider_data.reset_index(drop=True, inplace=True)
        guider_data.to_parquet(guider_path)

        # Write frame data.
        frame_data = gs.frame_data()
        frames_path = path.with_name(path.stem + "_frames.parquet")
        log.debug(f"Writing frame data to {frames_path!s}")
        frame_data.reset_index(drop=True, inplace=True)
        frame_data.to_parquet(frames_path)

        if generate_qa:
            log.info(f"Generating QA plots for {path.absolute()!s}")
            plot_qa(gs)

    return gs


def create_summary_file(
    files: Sequence[AnyPath],
    outpath: AnyPath,
    spec_framenos: list[int] | None = None,
):
    """Concatenates the guider data from a list of files.

    Parameters
    ----------
    files
        The list of files to concatenate. These must be the
        ``lvm.{telescope}.guider_{seqno}.parquet`` with one row corresponding
        to each guide step in the sequence.
    outpath
        The ``parquet`` table file where to save the data.
    spec_framenos
        A list of spectrograph frame numbers. The length must match
        the number of ``files``. If provided, a column ``spec_frameno`` will
        be added to each file table.

    Returns
    -------
    data
        A data frame with the concatenated data.

    """

    dfs: list[pandas.DataFrame] = []
    for ii, file in enumerate(files):
        df = pandas.read_parquet(file)

        if spec_framenos is not None and len(spec_framenos) > 0:
            df["spec_frameno"] = spec_framenos[ii]
            df["spec_frameno"] = df["spec_frameno"].astype("Int16")

        dfs.append(df)

    df_concat = pandas.concat(dfs)
    df_concat.sort_values(["telescope", "frameno"])

    outpath = pathlib.Path(outpath)
    outpath.parent.mkdir(parents=True, exist_ok=True)

    df_concat.reset_index(drop=True, inplace=True)
    df_concat.to_parquet(str(outpath))

    return df_concat


def coadd_camera(
    files: Sequence[AnyPath],
    outpath: str | None = "default",
    use_sigmaclip: bool = False,
    sigma: float = 3.0,
    database_profile: str = "default",
    database_params: dict = {},
    overwrite_reprocess: bool = False,
    multiprocess_mode: MULTIPROCESS_MODE_TYPE | None = None,
):
    """Co-adds a series of AG camera frames.

    This routine does the following:

    - Receives a range of guider frames that usually correspond to a full guider
      sequence.
    - For each frame in the range to co-add, substracts the dark current frame and
      derives counts per second.
    - Co-adds the frames using a sigma-clipped median.
    - Extracts sources from the image using SExtractor. Calculates PSF parameters
      by fitting Gaussian profiles to the x and y-axis marginal distributions.
    - Uses the first guide WCS to determine the centre of the image and query
      Gaia DR3 sources.
    - Uses a kd-tree to match Gaia objects to image sources.
    - Calculates image transparency as the median zero-point of an LVM-defined
      magnitude for each source. This AB magnitude is derived by integrating the
      Gaia XP mean spectrum over a bandpass with the effective response of the
      optical system and AG sensor.
    - Derives a new WCS solution for the co-added image by matching pixel positions
      for the sources in the co-added image to Gaia coordinates.
    - Writes the co-added image and metadata to disk.

    Parameters
    ----------
    files
        The list of files to be co-added. The must all correspond to a single
        camera and telescope.
    outpath
        The path of the co-added frame. If a relative path, it is written relative
        to the path of the first file to be co-added. If `None`, the file is not
        written to disk but the ``HDUList` is returned. If the value is
        ``"default"``, a default path is used.
    use_sigmaclip
        Whether to use sigma clipping when combining the stack of data. Disabled
        by default as it uses significant CPU and memory.
    sigma
        The sigma value for co-addition sigma clipping.
    database_profile
        Profile name to use to connect to the database and query Gaia.
    database_params
        Additional database parameters used to override the profile.
    overwrite_reprocess
        Whether to overwrite reprocessed files or use already existing ones.
    multiprocess_mode
        See `.process_all_spec_frames`.

    Returns
    -------
    camera_solution
        The `.CoAdd_CameraSolution` object holding the information about
        the co-added frame.

    """

    db_connection_params = {"profile": database_profile, **database_params}

    if outpath == "default":
        outpath = config["coadds"]["paths"]["coadd_camera_path"]

    # Get the list of frame numbers from the files.
    paths = sorted([pathlib.Path(file) for file in files])

    frame_nos = sorted([get_frameno(path) for path in paths])

    # Use the first file to get some common data (or at least it should be common!)
    sample_raw_header = get_raw_extension(paths[0]).header

    camname: str = sample_raw_header["CAMNAME"]
    telescope: str = sample_raw_header["TELESCOP"]
    dateobs0: str = sample_raw_header["DATE-OBS"]
    sjd = get_sjd("LCO", date=Time(dateobs0, format="isot").to_datetime())

    log.info(
        f"Co-adding frames {min(frame_nos)}-{max(frame_nos)} for MJD={sjd}, "
        f"camera={camname!r}, telescope={telescope!r}."
    )

    # Loop over each file, add the dark-subtracked data to the stack, and collect
    # frame metadata.
    create_framedata_partial = partial(
        create_framedata,
        db_connection_params=db_connection_params,
        overwrite_reprocess=overwrite_reprocess,
    )

    if multiprocess_mode == "frames":
        with multiprocessing.Pool(MULTIPROCESS_NCORES["frames"]) as pool:
            _frames = list(pool.map(create_framedata_partial, paths))
    else:
        _frames = list(map(create_framedata_partial, paths))

    frames = [frame for frame in _frames if frame is not None]

    data_stack: list[ARRAY_2D_F32] = []
    for fd in frames:
        if fd.data is not None and fd.stacked is True:
            data_stack.append(fd.data)
        fd.data = None

    coadd_image: ARRAY_2D_F32 | None = None
    coadd_solution: CoAdd_CameraSolution

    if len(data_stack) == 0 or telescope == "spec":
        if telescope == "spec":
            log.debug(f"Not stacking data for telecope {telescope!r}.")
        else:
            log.error(f"No data to stack for {telescope!r}.")

        coadd_solution = CoAdd_CameraSolution(
            frameno=-1,
            camera=camname,
            telescope=telescope,
        )

    else:
        log.debug(f"Creating median-combined co-added frame for {camname!r}.")

        # Combine the stack of data frames using the median of each pixel.
        # Optionally sigma-clip each pixel (this is computationally intensive).
        if use_sigmaclip:
            log.debug(f"Sigma-clipping stack with sigma={sigma}")
            stack_masked = sigma_clip(
                numpy.array(data_stack),
                sigma=int(sigma),
                maxiters=3,
                masked=True,
                axis=0,
                copy=False,
            )
            coadd_image = numpy.ma.median(stack_masked, axis=0).data

        else:
            coadd_image = numpy.median(numpy.array(data_stack), axis=0)

        del data_stack
        assert coadd_image is not None

        coadd_solution = process_camera_coadd(
            coadd_image,
            frames,
            db_connection_params=db_connection_params,
        )

    coadd_solution.frames = frames
    coadd_solution.sigmaclip = use_sigmaclip
    coadd_solution.sigmaclip_sigma = sigma if use_sigmaclip else None

    if outpath is not None:
        # Create the path for the output file.
        frameno0 = min(frame_nos)
        frameno1 = max(frame_nos)
        outpath = outpath.format(
            telescope=telescope,
            camname=camname,
            frameno0=frameno0,
            frameno1=frameno1,
            mjd=sjd,
        )

        if pathlib.Path(outpath).is_absolute():
            path = pathlib.Path(outpath)
        else:
            path = pathlib.Path(paths[0]).parent / outpath

        if path.parent.exists() is False:
            path.parent.mkdir(parents=True, exist_ok=True)

        coadd_solution.path = path

        log.info(f"Writing co-added frame to {path.absolute()!s}")

        # Write co-addded image and header.
        hdul = fits.HDUList(
            [
                fits.PrimaryHDU(),
                fits.CompImageHDU(
                    data=coadd_image,
                    header=create_coadd_header(coadd_solution),
                    name="COADD",
                ),
            ]
        )
        hdul.writeto(str(path), overwrite=True)

        # Write sources to file.
        if coadd_solution.sources is not None:
            sources_path = path.with_name(path.stem + "_sources.parquet")
            log.debug(f"Writing co-added sources to {sources_path!s}")
            coadd_solution.sources.reset_index(drop=True, inplace=True)
            coadd_solution.sources.to_parquet(sources_path)

        # Write frame data.
        frame_data = coadd_solution.frame_data()
        frames_path = path.with_name(path.stem + "_frames.parquet")
        log.debug(f"Writing frame data to {frames_path!s}")
        frame_data.reset_index(drop=True, inplace=True)
        frame_data.to_parquet(frames_path)

    return coadd_solution


def create_framedata(
    path: pathlib.Path,
    db_connection_params: dict = {},
    overwrite_reprocess: bool = False,
):
    """Collects information from a guider frame into a `.FrameData` object."""

    # Copy the original path.
    orig_path: pathlib.Path = path
    reprocessed: bool = False

    if not path.exists():
        log.error(f"Cannot find frame {path!s}")
        return None

    try:
        guiderv = fits.getval(str(path), "GUIDERV", "PROC")
    except Exception:
        guiderv = "0.0.0"

    if Version(guiderv) < Version("0.4.0a0"):
        new_path = reprocess_agcam(
            path,
            db_connection_params=db_connection_params,
            overwrite=overwrite_reprocess,
        )
        if new_path is not None:
            reprocessed = True
            path = new_path

    with fits.open(str(path)) as hdul:
        # RAW header
        raw_header = hdul["RAW"].header

        frameno = get_frameno(path)
        telescope = raw_header["TELESCOP"]
        camname = raw_header["CAMNAME"]
        date_obs = Time(raw_header["DATE-OBS"], format="isot")

        log_h = f"Frame {telescope}-{camname}-{frameno}:"

        # PROC header of the raw AG frame.
        if "PROC" not in hdul:
            log.error(f"{log_h} cannot find PROC extension.")
            return None

        proc_header = hdul["PROC"].header

        # We use the original path here always because reprocessed
        # files do not include image data.
        data, dark_sub = get_dark_subtracted_data(orig_path)
        if dark_sub is False:
            log.debug(f"{log_h} missing dark frame. Fitting and removing background.")

        wcs_mode = proc_header["WCSMODE"]
        stacked = (wcs_mode == "gaia") or reprocessed

        try:
            solution = CameraSolution.open(path)
        except ValueError as err:
            log.error(f"{log_h} failed creating CameraSolution ({err})")
            return None

        if solution.sources is None:
            log.warning(f"{log_h} no sources found.")

    # Add information as a FrameData.
    return FrameData(
        frameno=frameno,
        path=path,
        solution=solution,
        raw_header=dict(raw_header),
        camera=camname,
        telescope=telescope,
        date_obs=date_obs,
        sjd=get_sjd("LCO", date_obs.to_datetime()),
        exptime=raw_header["EXPTIME"],
        image_type=raw_header["IMAGETYP"],
        kmirror_drot=raw_header["KMIRDROT"],
        focusdt=raw_header["FOCUSDT"],
        stacked=stacked,
        data=data,
        reprocessed=reprocessed,
    )


def process_camera_coadd(
    data: ARRAY_2D_F32,
    framedata: list[FrameData],
    db_connection_params={},
):
    """Processes co-added data, extracts sources, and determines the WCS."""

    frame: FrameData | None = None
    for fd in framedata[::-1]:  # In reverse order. Last file is likely to be guiding.
        if fd.solution.solved:
            frame = fd
            break

    if frame is None or frame.solution.wcs is None:
        raise RuntimeError("Cannot find WCS reference solution for coadded frame.")

    # Extract sources in the co-added frame.
    coadd_sources = extract_marginal(
        data,
        box_size=31,
        threshold=3.0,
        max_detections=50,
        sextractor_quick_options={"minarea": 5},
    )
    coadd_sources["telescope"] = frame.telescope
    coadd_sources["camera"] = frame.camera

    # Add global frame pixels.
    xy = coadd_sources.loc[:, ["x", "y"]].to_numpy()
    ff_locs, _ = ag_to_full_frame(f"{frame.telescope}-{frame.camera[0]}", xy)
    coadd_sources.loc[:, ["x_ff", "z_ff"]] = ff_locs

    # Match with Gaia sources.
    coadd_sources, n_matches = match_with_gaia(
        frame.solution.wcs,
        coadd_sources,
        max_separation=3,
        concat=True,
        db_connection_params=db_connection_params,
    )

    wcs = None
    if n_matches < 5:
        log.warning("Insufficient number of Gaia matches. Cannot generate WCS.")
    else:
        # Get WCS solution from Gaia.
        wcs = wcs_from_gaia(coadd_sources)

    zp = estimate_zeropoint(data, coadd_sources)
    coadd_sources.update(zp)

    return CoAdd_CameraSolution(
        frameno=-1,
        coadd_image=data,
        telescope=frame.telescope,
        camera=frame.camera,
        sources=coadd_sources,
        wcs=wcs,
        wcs_mode="gaia",
        matched=wcs is not None,
    )


def get_guider_solutions(root: pathlib.Path, telescope: str, frameno: int):
    """Collects guider solutions for a frame."""

    filename = f"lvm.{telescope}.guider_{frameno:08d}.fits"

    log_h = f"Frame {telescope}-{frameno}:"

    path = root / filename
    if not path.exists():
        try:
            path = reprocess_legacy_guider_frame(root, frameno, telescope)
        except Exception:
            log.error(f"{log_h} failed retrieving guider solution.")
            return None

    guider_data = fits.getheader(path, "GUIDERDATA")

    solution = GuiderSolution.open(path)
    solution.guide_mode = guider_data["GUIDMODE"]
    solution.ra_field = guider_data["RAFIELD"]
    solution.dec_field = guider_data["DECFIELD"]
    solution.pa_field = guider_data.get("PAFIELD", numpy.nan)

    return solution


def create_coadd_header(solution: CoAdd_CameraSolution):
    """Creates the header object for a co-added frame."""

    assert solution.frames is not None
    assert solution.sources is not None

    telescope = solution.telescope

    frame_data = solution.frame_data()
    stacked = frame_data.loc[frame_data.stacked == 1, :]

    if len(frame_data) == 0 or len(stacked) == 0:
        raise ValueError("No stacked data found.")

    frame0 = frame_data.iloc[0].frameno
    framen = frame_data.iloc[-1].frameno
    stack0 = stacked.iloc[0].frameno
    stackn = stacked.iloc[-1].frameno

    isot = frame_data.iloc[0].date_obs
    sjd = get_sjd("LCO", Time(isot, format="isot").to_datetime())

    # Gets the range of FWHM in the sequence
    fwhm0: float | None = None
    fwhmn: float | None = None
    fwhm_median: float | None = None
    fwhm_medians = stacked.fwhm.dropna()
    if len(fwhm_medians) > 0:
        fwhm0 = fwhm_medians.iloc[0]
        fwhmn = fwhm_medians.iloc[-1]
        fwhm_median = float(numpy.median(fwhm_medians))

    # Gets the FWHM of the sources extracted from the stacked image.
    cofwhm = solution.fwhm
    cofwhmst = solution.sources.loc[solution.sources.valid == 1].fwhm.dropna().median()

    # Determine the PA drift due to k-mirror tracking.
    frame_pa = stacked.loc[:, ["frameno", "pa"]].dropna()
    pa_min: float = numpy.nan
    pa_max: float = numpy.nan
    pa_drift: float = numpy.nan

    pa_coeffs = numpy.array([numpy.nan, numpy.nan])
    drift_warn: bool = True

    if telescope == "spec":
        pa_min = frame_pa.pa.min()
        pa_max = frame_pa.pa.max()
        pa_drift = numpy.abs(pa_min - pa_max)
        drift_warn = False

    elif len(frame_pa) >= 2:
        pa_coeffs = polyfit_with_sigclip(
            frame_pa.frameno.to_numpy(numpy.float32),
            frame_pa.pa.to_numpy(numpy.float32),
            sigma=3,
            deg=1,
        )

        pa_vals = numpy.polyval(pa_coeffs, frame_pa.frameno)
        pa_min = numpy.round(pa_vals.min(), 4)
        pa_max = numpy.round(pa_vals.max(), 4)
        pa_drift = numpy.round(numpy.abs(pa_min - pa_max), 4)

        drift_warn = pa_drift > config["coadds"]["warnings"]["pa_drift"]

    wcs_header = solution.wcs.to_header() if solution.wcs is not None else []

    header = header_from_model("CAMERA_COADD")

    # Basic info
    header["TELESCOP"] = frame_data.iloc[0].telescope
    header["CAMNAME"] = frame_data.iloc[0].camera
    header["MJD"] = sjd
    header.insert("TELESCOP", ("", "/*** BASIC DATA ***/"))

    # Frame info
    header["FRAME0"] = frame0
    header["FRAMEN"] = framen
    header["NFRAMES"] = framen - frame0 + 1

    if telescope != "spec":
        header["STACK0"] = stack0
        header["STACKN"] = stackn
        header["NSTACKED"] = stackn - stack0 + 1
        header["COESTIM"] = "median"
        header["SIGCLIP"] = solution.sigmaclip
        header["SIGMA"] = solution.sigmaclip_sigma

    header["OBSTIME0"] = frame_data.iloc[0].date_obs
    header["OBSTIMEN"] = frame_data.iloc[-1].date_obs
    header["FWHM0"] = nan_or_none(fwhm0, 3)
    header["FWHMN"] = nan_or_none(fwhmn, 3)
    header["FWHMMED"] = nan_or_none(fwhm_median, 3)

    if telescope != "spec":
        header["COFWHM"] = nan_or_none(cofwhm, 3)
        header["COFWHMST"] = nan_or_none(cofwhmst, 3)

        header["PACOEFFA"] = nan_or_none(pa_coeffs[0])
        header["PACOEFFB"] = nan_or_none(pa_coeffs[1])

    header["PAMIN"] = nan_or_none(pa_min, 4)
    header["PAMAX"] = nan_or_none(pa_max, 4)
    header["PADRIFT"] = nan_or_none(pa_drift, 4)

    header["ZEROPT"] = nan_or_none(solution.zero_point, 3)

    header["SOLVED"] = solution.solved
    header.insert("FRAME0", ("", "/*** CO-ADDED PARAMETERS ***/"))

    # Warnings
    header["WARNPADR"] = drift_warn
    header["WARNTRAN"] = solution.transp_warning()
    header["WARNMATC"] = not solution.matched if telescope != "spec" else False
    header["WARNFWHM"] = solution.fwhm_warning()
    header.insert("WARNPADR", ("", "/*** WARNINGS ***/"))

    if solution.wcs is not None:
        header.extend(wcs_header)
        header.insert("WCSAXES", ("", "/*** CO-ADDED WCS ***/"))

    return header


def create_global_header(solution: GlobalSolution):
    """Creates the header object for a global frame.

    This function is very similar to `.create_coadd_header` but there are
    enough differences that it's worth keeping them separate and duplicate
    a bit of code.

    """

    assert solution.sources is not None

    telescope = solution.telescope

    frame_data = solution.frame_data()
    guider_data = solution.guider_data()

    if len(frame_data) == 0 or len(guider_data) == 0:
        raise RuntimeError("No frame or guider data to combine.")

    guiding = guider_data.loc[guider_data.guide_mode == "guide"]

    frame0 = frame_data.frameno.min()
    framen = frame_data.frameno.max()

    isot = frame_data.iloc[0].date_obs
    sjd = get_sjd("LCO", Time(isot, format="isot").to_datetime())

    # Gets the range of FWHM in the sequence
    fwhm0: float | None = None
    fwhmn: float | None = None
    fwhm_median: float | None = None
    fwhm_medians = guiding.fwhm.dropna()
    if len(fwhm_medians) > 0:
        fwhm0 = fwhm_medians.iloc[0]
        fwhmn = fwhm_medians.iloc[-1]
        fwhm_median = float(numpy.median(fwhm_medians))

    # Determine the PA drift due to k-mirror tracking.
    frame_pa = guiding.loc[:, ["frameno", "pa"]].dropna()
    pa_min: float = numpy.nan
    pa_max: float = numpy.nan
    pa_drift: float = numpy.nan

    pa_coeffs = numpy.array([numpy.nan, numpy.nan])
    drift_warn: bool = True

    if telescope == "spec":
        pa_min = frame_pa.pa.min()
        pa_max = frame_pa.pa.max()
        pa_drift = numpy.abs(pa_min - pa_max)
        drift_warn = False

    elif len(frame_pa) >= 2:
        pa_coeffs = polyfit_with_sigclip(
            frame_pa.frameno.to_numpy(numpy.float32),
            frame_pa.pa.to_numpy(numpy.float32),
            sigma=3,
            deg=1,
        )

        pa_vals = numpy.polyval(pa_coeffs, frame_pa.frameno)
        pa_min = numpy.round(pa_vals.min(), 4)
        pa_max = numpy.round(pa_vals.max(), 4)
        pa_drift = numpy.round(numpy.abs(pa_min - pa_max), 4)

        drift_warn = pa_drift > config["coadds"]["warnings"]["pa_drift"]

    wcs_header = []
    pa_warn: bool = False
    if solution.wcs is not None:
        wcs_header = solution.wcs.to_header()

        pa_field = guider_data.pa_field.iloc[0]
        pa = solution.pa
        pa_warn_threshold = config["coadds"]["warnings"]["pa_error"]

        if pa_field is not None:
            pa_warn = abs(angle_difference(pa, pa_field)) > pa_warn_threshold

    header = header_from_model("GLOBAL_COADD")

    # Basic info
    header["TELESCOP"] = frame_data.iloc[0].telescope
    header["MJD"] = sjd
    header.insert("TELESCOP", ("", "/*** BASIC DATA ***/"))

    # Frame info
    header["FRAME0"] = frame0
    header["FRAMEN"] = framen
    header["NFRAMES"] = framen - frame0 + 1

    header["OBSTIME0"] = frame_data.iloc[0].date_obs
    header["OBSTIMEN"] = frame_data.iloc[-1].date_obs

    header["FWHM0"] = nan_or_none(fwhm0, 3)
    header["FWHMN"] = nan_or_none(fwhmn, 3)
    header["FWHMMED"] = nan_or_none(fwhm_median, 3)

    if telescope != "spec":
        header["PACOEFFA"] = nan_or_none(pa_coeffs[0])
        header["PACOEFFB"] = nan_or_none(pa_coeffs[1])

    header["PAMIN"] = nan_or_none(pa_min, 4)
    header["PAMAX"] = nan_or_none(pa_max, 4)
    header["PADRIFT"] = nan_or_none(pa_drift, 4)

    header["ZEROPT"] = nan_or_none(solution.zero_point, 3)

    header["SOLVED"] = solution.solved
    header["NCAMSOL"] = solution.n_cameras_solved
    header.insert("FRAME0", ("", "/*** GLOBAL FRAME PARAMETERS ***/"))

    # Pointing
    if telescope != "spec":
        header["XFFPIX"] = guider_data.x_ff_pixel.iloc[0]
        header["ZFFPIX"] = guider_data.z_ff_pixel.iloc[0]
        header["RAFIELD"] = nan_or_none(guider_data.ra_field.iloc[0], 6)
        header["DECFIELD"] = nan_or_none(guider_data.dec_field.iloc[0], 6)
        header["PAFIELD"] = nan_or_none(guider_data.pa_field.iloc[0], 4)
        header["RAMEAS"] = nan_or_none(guider_data.ra.iloc[0], 6)
        header["DECMEAS"] = nan_or_none(guider_data.dec.iloc[0], 6)
        header["PAMEAS"] = nan_or_none(guider_data.pa.iloc[0], 4)

    # Warnings
    header["WARNPA"] = pa_warn
    header["WARNPADR"] = drift_warn
    header["WARNTRAN"] = solution.transp_warning()
    header["WARNMATC"] = solution.wcs is None if telescope != "spec" else False
    header["WARNFWHM"] = solution.fwhm_warning()
    header.insert("WARNPA", ("", "/*** WARNINGS ***/"))

    if solution.wcs is not None:
        header.extend(wcs_header)
        header.insert("WCSAXES", ("", "/*** CO-ADDED WCS ***/"))

    return header


def reprocess_agcam(file: AnyPath, overwrite: bool = False, db_connection_params={}):
    """Reprocesses an old file and converts it to the ``>=0.4.0`` format.

    .. warning::
        This function reprocesses a file on a best-effort basis. The astrometric
        solution is calculated using astrometry.net, which may be different
        from the original solution.

    Paremeters
    ----------
    file
        The file to reprocess.
    overwrite
        Exits early if the reprocessed file already exists.
    db_connection_params
        Database connection details to query Gaia.

    Returns
    -------
    processed
        The path to the processed file. The processed file is written relative
        to the parent of the input file as ``reprocessed/<file>``. Note that
        the reprocessed file does not include the raw original data to
        save space since that would not have changed. Reprocessed files are
        assigned ``GUIDERV=0.99.0``.

    """

    file = pathlib.Path(file)
    reproc_path = file.parent / "reprocessed" / file.name
    sources_path = reproc_path.with_suffix(".parquet")

    if reproc_path.exists() and not overwrite:
        log.debug(f"Found reprocessed file {reproc_path!s}")
        return reproc_path

    hdul_orig = fits.open(file)

    if "PROC" not in hdul_orig:
        # Just return, there's an informative error later.
        return

    log.warning(f"File {file!s} is too old and will be reprocessed.")

    if "RAW" in hdul_orig:
        raw = hdul_orig["RAW"].header
    else:
        raw = hdul_orig[0].header

    proc = hdul_orig["PROC"].header if "PROC" in hdul_orig else {}

    data_sub, _ = get_dark_subtracted_data(file)
    sources = extract_sources(file)

    solution: AstrometrySolution = solve_camera_with_astrometrynet(
        sources,
        ra=raw["RA"],
        dec=raw["DEC"],
    )
    wcs = solution.wcs

    # Now match with Gaia.
    if solution.wcs is not None:
        sources, _ = match_with_gaia(
            solution.wcs,
            sources,
            concat=True,
            db_connection_params=db_connection_params,
        )

        zp = estimate_zeropoint(data_sub, sources)
        sources.update(zp)

    darkfile = proc.get("DARKFILE", None)
    if darkfile:
        darkfile = os.path.basename(darkfile)

    new_proc = header_from_model("PROC")
    new_proc["TELESCOP"] = raw["TELESCOP"]
    new_proc["CAMNAME"] = raw["CAMNAME"]
    new_proc["DARKFILE"] = darkfile
    new_proc["DIRNAME"] = str(file.parent)
    new_proc["GUIDERV"] = "0.99.0"
    new_proc["SOURCESF"] = sources_path.name
    new_proc["PA"] = nan_or_none(get_crota2(wcs), 4) if wcs else None
    new_proc["ZEROPT"] = nan_or_none(sources.zp.dropna().median())
    new_proc["WCSMODE"] = "none" if wcs is None else "astrometrynet"
    new_proc["ORIGFILE"] = (file.name, "Original file name")
    new_proc["REPROC"] = (True, "Has this file been reprocessed?")

    if wcs is not None:
        new_proc.extend(wcs.to_header())

    reproc_path.parent.mkdir(parents=True, exist_ok=True)

    hdul = fits.HDUList(
        [
            fits.PrimaryHDU(),
            fits.ImageHDU(header=raw, name="RAW"),
            fits.ImageHDU(header=new_proc, name="PROC"),
        ]
    )
    hdul.writeto(str(reproc_path), overwrite=True)
    hdul.close()

    sources.reset_index(drop=True, inplace=True)
    sources.to_parquet(sources_path)

    return reproc_path


def reprocess_legacy_guider_frame(
    root: pathlib.Path,
    frameno: int,
    telescope: str,
    overwrite: bool = False,
):
    """Reprocesses a legacy ``proc-`` file.

    .. warning::
        This function reprocesses a file on a best-effort basis. Usually
        the individual frames used to generate the solution have also been
        reprocessed and the results may differ to those actually observed
        on sky.

    Paremeters
    ----------
    root
        The root path (usually the ``agcam`` directory for the given MJD).
    frameno
        The frame number.
    telescope
        The telescope that took the images.
    overwrite
        If `False` and a reprocessed ``lvm.guider`` file has been generated,
        uses that one.

    Returns
    -------
    processed
        The path to the processed file. The processed file is written relative
        to the parent of the input file as
        ``reprocessed/lvm,{telescope}.guider_{frameno:08d}.fits``. Reprocessed
        files are assigned ``GUIDERV=0.99.0``. The sources ``.parquet`` table
        file is also generated.


    """

    log_h = f"{telescope}-{frameno}:"

    proc_file = root / f"proc-lvm.{telescope}.agcam_{frameno:08d}.fits"
    guider_file = root / "reprocessed" / f"lvm.{telescope}.guider_{frameno:08d}.fits"
    guider_sources_file = guider_file.with_suffix(".parquet")

    if not proc_file.exists():
        raise FileNotFoundError(f"Cannot find file {proc_file!s}")

    if overwrite is False and guider_file.exists():
        log.debug(f"{log_h} found reprocessed lvm.guider file {guider_file!s}.")
        return guider_file

    log.warning(f"{log_h} using legacy proc- file {proc_file!s}.")

    proc = dict(fits.getheader(proc_file, "ASTROMETRY"))

    gdata_header = header_from_model("GUIDERDATA_HEADER")
    for key in gdata_header:
        if key in proc:
            gdata_header[key] = proc[key]

    gdata_header["GUIDERV"] = "0.99.0"
    gdata_header["DIRNAME"] = str(guider_file.parent)
    gdata_header["SOURCESF"] = guider_sources_file.name

    gdata_header["XFFPIX"] = config["xz_full_frame"][0]
    gdata_header["ZFFPIX"] = config["xz_full_frame"][1]

    gdata_header["GUIDMODE"] = "acquisition" if proc["ACQUISIT"] else "guide"

    # Get the sources. Try first the location of the new-style sources file.
    # Otherwise check if there's a reprocessed file.
    sources: list[pandas.DataFrame] = []
    n_solved: int = 0
    for key in ["FILE0", "FILE1"]:
        filex = proc.get(key, None)
        if filex is None:
            continue

        filex = pathlib.Path(filex).name

        found: bool = False
        if (root / filex).with_suffix(".parquet").exists():
            sources_file = (root / filex).with_suffix(".parquet")
            sources.append(pandas.read_parquet(sources_file))
            found = True

        elif (root / "reprocessed" / filex).with_suffix(".parquet").exists():
            sources_file = (root / "reprocessed" / filex).with_suffix(".parquet")
            sources.append(pandas.read_parquet(sources_file))
            found = True

        if found:
            if "east" in filex:
                gdata_header["FILEEAST"] = filex
            else:
                gdata_header["FILEWEST"] = filex

            hdul = fits.open(root / filex)
            if "PROC" in hdul and hdul["PROC"].header["WCSMODE"] != "none":
                n_solved += 1

            hdul.close()

    if len(sources) == 0:
        raise RuntimeError(f"No sources found for guider frame {proc_file!s}")

    gdata_header["NCAMSOL"] = n_solved

    sources_concat = pandas.concat(sources)
    if len(sources_concat.ra.dropna()) > 5:
        wcs = wcs_from_gaia(sources_concat, xy_cols=["x_ff", "z_ff"])
    else:
        log.warning(f"Insufficient Gaia matches for {proc_file!s}. Cannot fit WCS.")
        wcs = None

    if wcs is not None:
        gdata_header.extend(wcs.to_header())

        gdata_header["PAMEAS"] = nan_or_none(get_crota2(wcs), 4)
        gdata_header["ZEROPT"] = nan_or_none(sources_concat.zp.dropna().median(), 3)
        gdata_header["SOLVED"] = True

    hdul = fits.HDUList(
        [
            fits.PrimaryHDU(),
            fits.ImageHDU(header=gdata_header, name="GUIDERDATA"),
        ]
    )
    hdul.writeto(str(guider_file), overwrite=True)

    sources_concat.reset_index(drop=True, inplace=True)
    sources_concat.to_parquet(guider_sources_file)

    return guider_file


def watch_for_files(base_path: str = "/data/spectro"):
    """Watches and processes guider frames for new spectrograph images."""

    AG_PATH = pathlib.Path("/data/agcam")

    sjd: int = 0

    observer = PollingObserver(timeout=5)
    handler = SpecPatternEventHandler()

    observer.schedule(handler, "/data/agcam")
    observer.start()

    try:
        while True:
            time.sleep(1)

            # Check if the SJD has changed.
            new_sjd = get_sjd("LCO")

            if new_sjd != sjd:
                observer.unschedule_all()

                if not os.path.exists(f"{base_path}/{new_sjd}"):
                    continue

                log.info(f"Switching to SJD {new_sjd}")
                sjd = new_sjd

                (AG_PATH / str(sjd) / "coadds").mkdir(parents=True, exist_ok=True)
                log.start_file_logger(
                    str(AG_PATH / f"{sjd}/coadds/coadds_{sjd}.log"),
                    rotating=False,
                )

                observer.schedule(handler, f"{base_path}/{sjd}")

                log.info(f"Watching directory {base_path}/{sjd}")

    except KeyboardInterrupt:
        observer.stop()
    except Exception as err:
        print(err)

    observer.join()


class SpecPatternEventHandler(PatternMatchingEventHandler):
    """Handles newly created spectrograph files."""

    def __init__(self):
        self.lock = Lock()

        super().__init__(
            patterns=["*-b1-*.fits.gz"],
            ignore_directories=True,
            case_sensitive=True,
        )

    def on_created(self, event: FileCreatedEvent):
        """Runs the co-add code when a new file is created."""

        # Do not process more than one file at the same time.
        while self.lock.locked():
            time.sleep(1)

        self.lock.acquire()

        try:
            new_file = event.src_path

            if new_file is None or new_file == "":
                return

            path = pathlib.Path(new_file).absolute()

            if not path.exists():
                log.warning(f"Detected file {path!s} does not exist!")
                return

            image_type = fits.getval(str(path), "IMAGETYP")
            exptime = fits.getval(str(path), "EXPTIME")
            if image_type != "object" or exptime < 60:
                log.info(f"Detected file {path}. Not an object image, skipping.")
                return

            log.info(f"Processing spectrograph frame {get_spec_frameno(path)}")
            outpaths = coadd_from_spec_frame(path, multiprocess_mode="telescopes")

            if len(outpaths) == 0:
                log.info(f"All done for {path}")
                return

            # Recreate the summary files.
            log.info("Updating summary files.")

            sjd = get_sjd("LCO")
            parent_coadds = outpaths[0].parent

            for table in ["frames", "guiderdata"]:
                cfs = parent_coadds.glob("lvm.*.coadd*.fits")
                tfs = [pp.with_name(pp.stem + f"_{table}.parquet") for pp in cfs]
                spec_nos = [int(ff.name.split("_")[-2][1:]) for ff in tfs]

                if len(spec_nos) > 0:
                    create_summary_file(
                        tfs,
                        parent_coadds / f"lvm.guider.mjd_{sjd}_{table}.parquet",
                        spec_framenos=spec_nos,
                    )

            log.info(f"All done for {path}")

        except Exception as err:
            log.error(f"Error found while processing file: {err}")

        finally:
            # Release the lock so other files can be processed.
            self.lock.release()


def coadd_to_database(
    coadd_file: AnyPath,
    exposure_no: int | None = None,
    **db_connection_params,
):
    """Loads the co-add data into the databse."""

    coadd_file = pathlib.Path(coadd_file)

    log.info("Ingesting co-added files into the database.")

    if not coadd_file.exists():
        raise RuntimeError(f"Cannot find co-add file {coadd_file}")

    global_h = fits.getheader(coadd_file, "GLOBAL")
    sjd = global_h["MJD"]

    delete_columns: list[str] = ["telescope"]
    if exposure_no is not None:
        delete_columns.append("exposure_no")

    # Load global coadd data.
    log.debug("Loading coadd data to database.")
    cards = [
        "mjd",
        "frame0",
        "framen",
        "nframes",
        "obstime0",
        "obstimen",
        "fwhm0",
        "fwhmn",
        "fwhmmed",
        "pacoeffa",
        "pacoeffb",
        "pamin",
        "pamax",
        "padrift",
        "zeropt",
        "solved",
        "ncamsol",
        "xffpix",
        "zffpix",
        "rafield",
        "decfield",
        "pafield",
        "rameas",
        "decmeas",
        "pameas",
        "warnpa",
        "warnpadr",
        "warntran",
        "warnmatc",
        "warnfwhm",
    ]
    try:
        global_df = pandas.DataFrame()
        for card in cards:
            global_df[card] = [global_h[card]]
        global_df["telescope"] = global_h["TELESCOP"]
        if exposure_no:
            global_df["exposure_no"] = exposure_no

        table_name = config["database"]["guider_coadd_table"]
        dataframe_to_database(
            global_df,
            table_name,
            delete_columns=delete_columns,
            **db_connection_params,
        )
    except Exception as err:
        log.error(f"Failed loading frames to database: {str(err).strip()}")

    # Load AG frames.
    ag_frames_file = coadd_file.with_name(coadd_file.stem + "_frames.parquet")
    if ag_frames_file.exists():
        log.debug("Loading AG frames to database.")
        try:
            ag_frames = pandas.read_parquet(ag_frames_file)
            if exposure_no:
                ag_frames["exposure_no"] = exposure_no
            if len(ag_frames) > 0:
                ag_frames_mjd = isot_to_sjd(ag_frames["date_obs"].iloc[0])
                ag_frames["mjd"] = ag_frames_mjd
                ag_frames["stacked"] = ag_frames.stacked.astype(bool)
                ag_frames["solved"] = ag_frames.solved.astype(bool)
                table_name = config["database"]["agcam_frame_table"]
                dataframe_to_database(
                    ag_frames,
                    table_name,
                    delete_columns=delete_columns,
                    **db_connection_params,
                )
        except Exception as err:
            log.error(f"Failed loading frames to database: {str(err).strip()}")
    else:
        log.error(f"Cannot find AG frames file {ag_frames_file}")

    # Load guider data.
    guider_frames_file = coadd_file.with_name(coadd_file.stem + "_guiderdata.parquet")
    if guider_frames_file.exists():
        log.debug("Loading guider frames to database.")
        try:
            guider_frames = pandas.read_parquet(guider_frames_file)
            if exposure_no:
                guider_frames["exposure_no"] = exposure_no
            if len(guider_frames) > 0:
                guider_frames["mjd"] = sjd
                guider_frames["solved"] = guider_frames.solved.astype(bool)
                guider_frames["applied"] = guider_frames.applied.astype(bool)
                table_name = config["database"]["guider_frame_table"]
                dataframe_to_database(
                    guider_frames,
                    table_name,
                    delete_columns=delete_columns,
                    **db_connection_params,
                )
        except Exception as err:
            log.error(f"Failed loading frames to database: {str(err).strip()}")
    else:
        log.error(f"Cannot find guider frames file {guider_frames_file}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-30
# @Filename: types.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import nptyping as npt


# Array types
ARRAY_2D_U16 = npt.NDArray[npt.Shape["*, *"], npt.UInt16]
ARRAY_2D_F32 = npt.NDArray[npt.Shape["*, *"], npt.Float32]
ARRAY_2D_F64 = npt.NDArray[npt.Shape["*, *"], npt.Float64]

ARRAY_1D_F32 = npt.NDArray[npt.Shape["*"], npt.Float32]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-30
# @Filename: dataclasses.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
import warnings
from dataclasses import dataclass, field

from typing import Literal

import nptyping as npt
import numpy
import pandas
from astropy.io import fits
from astropy.time import Time
from astropy.wcs import WCS, FITSFixedWarning
from packaging.version import Version

from lvmguider import config, log
from lvmguider.tools import dataframe_from_model, get_frameno
from lvmguider.transformations import get_crota2
from lvmguider.types import ARRAY_2D_F32


__all__ = ["CameraSolution", "GuiderSolution", "FrameData"]


WCS_MODE_T = Literal["none"] | Literal["astrometrynet"] | Literal["gaia"]


@dataclass(kw_only=True)
class BaseSolution:
    """A base class for an object with an astrometric solution."""

    telescope: str
    wcs: WCS | None = None
    matched: bool = False

    def __post_init__(self):
        if self.sources is None:
            self.sources = dataframe_from_model("SOURCES")

    @property
    def solved(self):
        """Was the frame solved?"""

        return self.wcs is not None

    @property
    def pa(self):
        """Returns the position angle from the WCS."""

        if self.wcs is not None:
            return get_crota2(self.wcs)

        return numpy.nan

    @property
    def zero_point(self):
        """Returns the median zero point."""

        if self.sources is not None:
            return float(self.sources.zp.dropna().median())

        return numpy.nan

    @property
    def fwhm(self):
        """Returns the FWHM from the extracted sources."""

        if self.sources is not None:
            return self.sources.loc[self.sources.valid == 1].fwhm.dropna().median()

        return numpy.nan

    @property
    def pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the camera pointing at its central pixel."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*config["xz_ag_frame"])
        return numpy.array([skyc.ra.deg, skyc.dec.deg])


@dataclass(kw_only=True)
class CameraSolution(BaseSolution):
    """A camera solution, including the determined WCS."""

    frameno: int
    camera: str
    path: pathlib.Path
    sources: pandas.DataFrame | None = None
    ref_frame: pathlib.Path | None = None
    wcs_mode: WCS_MODE_T = "none"
    guider_version: Version = Version("0.0.0")

    @classmethod
    def open(cls, file: str | pathlib.Path):
        """Creates an instance from an ``lvm.agcam`` file."""

        file = pathlib.Path(file)
        hdul = fits.open(str(file))

        if "PROC" not in hdul:
            raise ValueError("HDU list does not have a PROC extension.")

        raw = hdul["RAW"].header
        proc = hdul["PROC"].header

        if "GUIDERV" not in proc:
            raise ValueError("Guider version not found.")

        guiderv = Version(proc["GUIDERV"])
        if guiderv < Version("0.4.0a0"):
            raise ValueError(
                "The file was generated with an unsupported version of lvmguider."
            )

        sources: pandas.DataFrame | None = None
        if proc["SOURCESF"]:
            dirname = file.parent
            sources = pandas.read_parquet(dirname / proc["SOURCESF"])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FITSFixedWarning)
            wcs_mode = proc["WCSMODE"]
            wcs = WCS(proc) if wcs_mode != "none" else None

            # Fix cases when WCSMODE is not "none" but the WCS is not valid.
            if wcs is not None and not wcs.is_celestial:
                log.warning(f"Invalid WCS found for {file!s}.")
                wcs = None

        ref_file = proc["REFFILE"]
        ref_frame = pathlib.Path(ref_file) if ref_file is not None else None

        return CameraSolution(
            frameno=get_frameno(file),
            telescope=raw["TELESCOP"],
            camera=raw["CAMNAME"],
            path=file.absolute(),
            sources=sources,
            wcs=wcs,
            wcs_mode=wcs_mode,
            matched=len(sources.ra.dropna()) > 0 if sources is not None else False,
            ref_frame=ref_frame,
            guider_version=guiderv,
        )


@dataclass(kw_only=True)
class GuiderSolution(BaseSolution):
    """A class to hold an astrometric solution determined by the guider.

    This class abstracts an astrometric solution regardless of whether it was
    determined using astrometry.net or Gaia sources.

    """

    frameno: int
    telescope: str
    solutions: list[CameraSolution]
    guide_pixel: npt.NDArray[npt.Shape["2"], npt.Float32]
    ra_field: float = numpy.nan
    dec_field: float = numpy.nan
    pa_field: float = numpy.nan
    ra_off: float = numpy.nan
    dec_off: float = numpy.nan
    pa_off: float = numpy.nan
    axis0_off: float = numpy.nan
    axis1_off: float = numpy.nan
    correction_applied: bool = False
    correction: list[float] = field(default_factory=lambda: [0.0, 0.0, 0.0])
    guide_mode: str = "guide"
    guider_version: Version = Version("0.0.0")

    @property
    def sources(self):
        """Concatenates sources from the co-added solutions."""

        return pandas.concat(
            [cs.sources for cs in self.solutions if cs.sources is not None],
            axis=0,
        )

    def __getitem__(self, key: str):
        """Returns the associated camera solution."""

        for solution in self.solutions:
            if solution.camera == key:
                return solution

        raise KeyError(f"Invalid camera name {key!r}.")

    @property
    def pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the telescope pointing at boresight from the WCS."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*config["xz_full_frame"])
        return numpy.array([skyc.ra.deg, skyc.dec.deg])

    @property
    def pixel_pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the telescope pointing at the full frame pixel from the WCS."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*self.guide_pixel)
        return numpy.array([skyc.ra.deg, skyc.dec.deg])

    @property
    def solved(self):
        """Was the frame solved?"""

        return self.wcs is not None

    @property
    def cameras(self):
        """Returns a list of cameras."""

        return [cs.camera for cs in self.solutions]

    @property
    def separation(self):
        """Returns the separation between field centre and telescope pointing."""

        return numpy.hypot(self.ra_off, self.dec_off)

    @property
    def n_cameras_solved(self):
        """Returns the number of cameras solved."""

        return len([cs for cs in self.solutions if cs.solved])

    def all_cameras_solved(self):
        """Returns `True` if all the cameras have solved."""

        return all([cs.solved for cs in self.solutions])

    @classmethod
    def open(cls, file: str | pathlib.Path, dirname: str | pathlib.Path | None = None):
        """Creates an instance from an ``lvm.guider`` file."""

        file = pathlib.Path(file)
        hdul = fits.open(str(file))

        guider_data = hdul["GUIDERDATA"].header

        if "GUIDERV" not in guider_data:
            raise ValueError("Guider version not found.")

        guiderv = Version(guider_data["GUIDERV"])
        if guiderv < Version("0.4.0a0"):
            raise ValueError(
                "The file was generated with an unsupported version of lvmguider."
            )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FITSFixedWarning)
            wcs = WCS(guider_data) if guider_data["SOLVED"] else None

        solutions: list[CameraSolution] = []
        for key in ["FILEEAST", "FILEWEST"]:
            if guider_data[key] is not None:
                dirname = pathlib.Path(dirname or guider_data["DIRNAME"] or file.parent)
                solutions.append(CameraSolution.open(dirname / guider_data[key]))

        if "XFFPIX" in guider_data:
            guide_pixel = numpy.array([guider_data["XFFPIX"], guider_data["ZFFPIX"]])
        elif "XMFPIX" in guider_data:
            # From when we called it master frame.
            guide_pixel = numpy.array([guider_data["XMFPIX"], guider_data["ZMFPIX"]])
        else:
            log.warning("Missing pixel information. Default to central pixel.")
            guide_pixel = config["xz_full_frame"]

        return GuiderSolution(
            frameno=get_frameno(file),
            telescope=guider_data["TELESCOP"],
            solutions=solutions,
            guide_pixel=guide_pixel,
            wcs=wcs,
            ra_off=guider_data["OFFRAMEA"] or numpy.nan,
            dec_off=guider_data["OFFDEMEA"] or numpy.nan,
            pa_off=guider_data["OFFPAMEA"] or numpy.nan,
            axis0_off=guider_data["OFFA0MEA"] or numpy.nan,
            axis1_off=guider_data["OFFA1MEA"] or numpy.nan,
            correction_applied=guider_data["CORRAPPL"],
            correction=[
                guider_data["AX0CORR"],
                guider_data["AX1CORR"],
                guider_data["PACORR"],
            ],
            guider_version=guiderv,
        )


@dataclass(kw_only=True)
class FrameData:
    """Data associated with a frame."""

    frameno: int
    path: pathlib.Path
    raw_header: dict
    date_obs: Time
    sjd: int
    camera: str
    telescope: str
    exptime: float
    image_type: str
    kmirror_drot: float
    focusdt: float
    solution: CameraSolution
    stacked: bool = False
    reprocessed: bool = False
    data: ARRAY_2D_F32 | None = None

    @property
    def sources(self):
        """Accessor to the camera solution sources."""

        return self.solution.sources


class CoAddWarningsMixIn:
    """Methods to calculate warnings for co-added data classes."""

    zero_point: float
    fwhm: float
    sources: pandas.DataFrame

    def transp_warning(self):
        """Determines whether to raise a transparency warning."""

        # We raise a transparency warning when the median ZP is > some nominal plus
        # a number of magnitudes. The nominal ZP corresponds to a photometric night.
        nominal_zp = config["coadds"]["warnings"]["nominal_zp"]
        zp_overmag_warning = config["coadds"]["warnings"]["zp_overmag_warning"]

        if numpy.isnan(self.zero_point):
            return True
        else:
            return self.zero_point > (nominal_zp + zp_overmag_warning)

    def fwhm_warning(self):
        """Determines whether to raise a transparency warning."""

        # Similarly, for FWHM we raise a warning if the co-added FWHM
        # is > factor * median FWHM. This may indicate that there was a jump
        # during guiding.
        if numpy.isnan(self.zero_point):
            return True
        else:
            fwhm_factor_warn = config["coadds"]["warnings"]["fwhm_factor_warning"]
            sources_fwhm = self.sources.loc[self.sources.valid == 1, "fwhm"]
            return self.fwhm > sources_fwhm.dropna().median() * fwhm_factor_warn


@dataclass(kw_only=True)
class CoAdd_CameraSolution(CameraSolution, CoAddWarningsMixIn):
    """A camera solution for a co-added frame."""

    path: pathlib.Path = pathlib.Path("")
    coadd_image: ARRAY_2D_F32 | None = None
    frames: list[FrameData] | None = None
    sources: pandas.DataFrame | None = None
    sigmaclip: bool = False
    sigmaclip_sigma: float | None = None

    def frame_data(self):
        """Returns a Pandas data frame from a list of `.FrameData`."""

        df = dataframe_from_model("FRAMEDATA")

        if self.frames is None:
            return df

        for ii, fd in enumerate(self.frames):
            new_row = dict(
                frameno=fd.frameno,
                date_obs=fd.date_obs.isot,
                camera=fd.camera,
                telescope=fd.telescope,
                exptime=numpy.float32(fd.exptime),
                kmirror_drot=numpy.float32(fd.kmirror_drot),
                focusdt=numpy.float32(fd.focusdt),
                fwhm=numpy.float32(fd.solution.fwhm),
                pa=numpy.float32(fd.solution.pa),
                zero_point=numpy.float32(fd.solution.zero_point),
                stacked=int(fd.stacked),
                solved=int(fd.solution.solved),
                wcs_mode=fd.solution.wcs_mode,
            )
            df.loc[ii, list(new_row)] = list(new_row.values())

        df = df.sort_values(["frameno", "camera"])

        df.reset_index(drop=True, inplace=True)

        return df


@dataclass(kw_only=True)
class GlobalSolution(BaseSolution, CoAddWarningsMixIn):
    """A global solution from co-added frames."""

    coadd_solutions: list[CoAdd_CameraSolution]
    guider_solutions: list[GuiderSolution]
    path: pathlib.Path = pathlib.Path("")
    matched: bool = True

    @property
    def n_cameras_solved(self):
        """Returns the number of cameras solved."""

        return len([cs for cs in self.coadd_solutions if cs.solved])

    @property
    def sources(self):
        """Concatenates sources from the co-added solutions."""

        return pandas.concat(
            [cs.sources for cs in self.coadd_solutions if cs.sources is not None],
            axis=0,
        )

    @property
    def pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the camera pointing at its central pixel."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*config["xz_full_frame"])
        return numpy.array([skyc.ra.deg, skyc.dec.deg])

    def frame_data(self):
        """Concatenates the frame data and returns a Pandas data frame."""

        frame_data = pandas.concat([cs.frame_data() for cs in self.coadd_solutions])
        frame_data.sort_values(["frameno", "camera"], inplace=True)

        return frame_data

    def guider_data(self):
        """Constructs a data frame with guider data."""

        df = dataframe_from_model("GUIDERDATA_FRAME")

        for ii, gs in enumerate(self.guider_solutions):
            pa = get_crota2(gs.wcs) if gs.wcs is not None else numpy.nan
            pointing = gs.pointing

            new_row = dict(
                frameno=gs.frameno,
                telescope=gs.telescope,
                fwhm=numpy.float32(gs.fwhm),
                pa=numpy.float32(pa),
                zero_point=numpy.float32(gs.zero_point),
                solved=int(gs.solved),
                n_cameras_solved=int(gs.n_cameras_solved),
                guide_mode=gs.guide_mode,
                x_ff_pixel=numpy.float32(gs.guide_pixel[0]),
                z_ff_pixel=numpy.float32(gs.guide_pixel[1]),
                ra=numpy.float64(pointing[0]),
                dec=numpy.float64(pointing[1]),
                ra_field=numpy.float64(gs.ra_field),
                dec_field=numpy.float64(gs.dec_field),
                pa_field=numpy.float32(gs.pa_field),
                ra_off=numpy.float32(gs.ra_off),
                dec_off=numpy.float32(gs.dec_off),
                pa_off=numpy.float32(gs.pa_off),
                axis0_off=numpy.float32(gs.axis0_off),
                axis1_off=numpy.float32(gs.axis1_off),
                applied=int(gs.correction_applied),
                ax0_applied=numpy.float32(gs.correction[0]),
                ax1_applied=numpy.float32(gs.correction[1]),
                rot_applied=numpy.float32(gs.correction[2]),
            )
            df.loc[ii, list(new_row)] = list(new_row.values())

        df = df.sort_values(["frameno"])
        df.reset_index(drop=True, inplace=True)

        return df
